import ast
import os
import re
import numpy as np
import tifffile


def get_default(x, default):
    return default if x is None else x


def ensure_list(x) -> list:
    if isinstance(x, list):
        return x
    else:
        return [x]


def filter_dict(dict0):
    new_dict = {}
    for key, value0 in dict0.items():
        if value0 is not None:
            values = []
            for value in ensure_list(value0):
                if isinstance(value, dict):
                    value = filter_dict(value)
                values.append(value)
            if len(values) == 1:
                values = values[0]
            new_dict[key] = values
    return new_dict


def tags_to_dict(tags: tifffile.TiffTags) -> dict:
    tag_dict = {}
    for tag in tags.values():
        tag_dict[tag.name] = tag.value
    return tag_dict


def desc_to_dict(desc: str) -> dict:
    desc_dict = {}
    if desc.startswith('{'):
        try:
            metadata = ast.literal_eval(desc)
            return metadata
        except:
            pass
    for item in re.split(r'[\r\n\t|]', desc):
        item_sep = '='
        if ':' in item:
            item_sep = ':'
        if item_sep in item:
            items = item.split(item_sep)
            key = items[0].strip()
            value = items[1].strip()
            try:
                if '.' in value:
                    value = float(value)
                else:
                    value = int(value)
            except:
                try:
                    value = bool(value)
                except:
                    pass
            desc_dict[key] = value
    return desc_dict


def print_dict(dct: dict, compact: bool = False, indent: int = 0):
    s = ''
    for key, value in dct.items():
        if not isinstance(value, list):
            if not compact: s += '\t' * indent
            s += str(key)
            s += ':' if compact else '\n'
        if isinstance(value, dict):
            s += print_dict(value, indent=indent + 1)
        elif isinstance(value, list):
            for v in value:
                s += print_dict(v)
        else:
            if not compact: s += '\t' * (indent + 1)
            s += str(value)
        s += ' ' if compact else '\n'
    return s


def print_hbytes(bytes: int) -> str:
    exps = ['', 'K', 'M', 'G', 'T']
    div = 1024
    exp = 0

    while bytes > div:
        bytes /= div
        exp +=1
    return f'{bytes:.1f}{exps[exp]}B'


def check_round_significants(a: float, significant_digits: int) -> float:
    rounded = round_significants(a, significant_digits)
    if a != 0:
        dif = 1 - rounded / a
    else:
        dif = rounded - a
    if abs(dif) < 10 ** -significant_digits:
        return rounded
    return a


def round_significants(a: float, significant_digits: int) -> float:
    if a != 0:
        round_decimals = significant_digits - int(np.floor(np.log10(abs(a)))) - 1
        return round(a, round_decimals)
    return a


def get_filetitle(filename: str, remove_all_ext: bool = False) -> str:
    filebase = os.path.basename(filename)
    if remove_all_ext:
        return filebase.split('.')[0]
    else:
        return os.path.splitext(filebase)[0]


def split_num_text(text):
    num_texts = []
    block = ''
    is_num0 = None
    if text is None:
        return None

    for c in text:
        is_num = (c.isnumeric() or c == '.')
        if is_num0 is not None and is_num != is_num0:
            num_texts.append(block)
            block = ''
        block += c
        is_num0 = is_num
    if block != '':
        num_texts.append(block)

    num_texts2 = []
    for block in num_texts:
        block = block.strip()
        try:
            block = float(block)
        except:
            pass
        if block not in [' ', ',', '|']:
            num_texts2.append(block)
    return num_texts2


def split_value_unit_list(text):
    value_units = []
    if text is None:
        return None

    items = split_num_text(text)
    if isinstance(items[-1], str):
        def_unit = items[-1]
    else:
        def_unit = ''

    i = 0
    while i < len(items):
        value = items[i]
        if i + 1 < len(items):
            unit = items[i + 1]
        else:
            unit = ''
        if not isinstance(value, str):
            if isinstance(unit, str):
                i += 1
            else:
                unit = def_unit
            value_units.append((value, unit))
        i += 1
    return value_units


def get_value_units_micrometer(value_units0: list):
    conversions = {'nm': 1e-3, 'µm': 1, 'um': 1, 'mm': 1e3, 'cm': 1e4, 'm': 1e6}
    if value_units0 is None:
        return None

    value_units = [value_unit[0] * conversions.get(value_unit[1], 1) for value_unit in value_units0]
    return value_units



from __future__ import annotations
import logging
import os
import pandas as pd
from types import TracebackType

from Omero import Omero


class OmeroLabelReader:
    """Omero metadata extraction to label file"""

    params: dict
    """input parameters"""
    omero: Omero
    """Omero instance"""
    manage_omero: bool
    """If responsible for managing Omero instance"""

    def __init__(self, params: dict, omero: Omero = None):
        self.params = params
        self.manage_omero = (omero is None)
        if self.manage_omero:
            self.omero = Omero(params)
        else:
            self.omero = omero

    def __enter__(self) -> OmeroLabelReader:
        if self.manage_omero:
            self.omero.init()
        return self

    def __exit__(self, exc_type: type[BaseException], exc_value: BaseException, traceback: TracebackType):
        if self.manage_omero:
            self.omero.close()

    def create_label_csv(self, image_ids):
        image_names = []
        image_annotations = []
        input = self.params['input']
        output = self.params['output']
        input_labels = input.get('omero', {}).get('labels', [])
        logging.info(f'Matching images: {len(image_ids)}')
        for id in image_ids:
            name, annotations = self.omero.get_image_annotation(id, input_labels)
            image_names.append(name)
            image_annotations.append(annotations)
        df = pd.DataFrame(index=image_ids, data=image_annotations)
        df.index.name = 'omero_id'
        for input_label in input_labels:
            if input_label in df:
                logging.info(f'Label {input_label}:\n' + df[input_label].value_counts().to_string())
        df.insert(0, 'omero_name', image_names)
        df['path'] = [image_name + '.' + output['format'] for image_name in image_names]
        log_path = os.path.dirname(output['csv'])
        if not os.path.exists(log_path):
            os.makedirs(log_path)
        df.to_csv(output['csv'])



from __future__ import annotations
import xmltodict


class XmlDict(dict):
    """xmltodict type dictionary providing and propagating access to keys without @ sign"""

    def __getitem__(self, key: str):
        at_key = '@' + key
        if key not in self and at_key in self:
            key = at_key
        value = dict.__getitem__(self, key)
        if isinstance(value, dict):
            value = XmlDict(value)
        if isinstance(value, list):
            value = XmlList(value)
        return value

    def get(self, key: str, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def copy(self) -> XmlDict:
        return XmlDict(dict.copy(self))


class XmlList(list):
    """xmltodict type list propagating access to keys without @ sign"""

    def __getitem__(self, index: int):
        value = list.__getitem__(self, index)
        if isinstance(value, dict):
            value = XmlDict(value)
        if isinstance(value, list):
            value = XmlList(value)
        return value

    def __iter__(self) -> XmlList:
        self.i = 0
        return self

    def __next__(self):
        item = self[self.i]
        self.i += 1
        if self.i < len(self):
            return item
        else:
            raise StopIteration


def xml2dict(xml_metadata: str) -> dict:
    return XmlDict(xmltodict.parse(xml_metadata))


def dict2xml(dct: dict) -> str:
    return xmltodict.unparse(dct, short_empty_elements=True, pretty=True)


if __name__ == '__main__':
    dct = {'@a': 1, 'b': {'@aa': 2, 'bb': 3}}
    test = XmlDict(dct)
    print(test.get('a', 11))
    print(test['a'])
    print(test.get('b').get('aa'))
    print(test['b']['aa'])

    xml = xmltodict.unparse({'TEST': test}, short_empty_elements=True, pretty=True)
    print(xml)



from __future__ import annotations
import logging
import omero
from omero.gateway import BlitzGateway
import omero.model
from types import TracebackType

from src.OmeSliCC.omero_credentials import decrypt_credentials
from src.OmeSliCC.util import *


class Omero:
    """Omero image and metadata extraction"""
    
    def __init__(self, params: dict):
        self.params = params
        self.private_key_filename = params['credentials']['private_key']
        self.credentials_filename = params['credentials']['credentials']
        self.connected = False

    def __enter__(self) -> Omero:
        self.init()
        return self

    def __exit__(self, exc_type: type[BaseException], exc_value: BaseException, traceback: TracebackType):
        self.close()

    def init(self):
        self._connect()
        self._switch_user_group()

    def close(self):
        self._disconnect()

    def _connect(self):
        logging.info('Connecting to Omero...')
        usr, pwd = decrypt_credentials(self.private_key_filename, self.credentials_filename)
        self.conn = BlitzGateway(usr, pwd, host=self.params['input']['omero']['host'], secure=True)
        if not self.conn.connect():
            self._disconnect()
            logging.error('Omero connection error')
            raise ConnectionError
        self.conn.c.enableKeepAlive(60)
        self.connected = True
        logging.info(f'Connected as {self.conn.getUser().getName()}')

    def _disconnect(self):
        self.conn.close()
        self.connected = False

    def _switch_user_group(self):
        self.conn.SERVICE_OPTS.setOmeroGroup('-1')

    def _get_project(self, project_id: int) -> omero.gateway.ProjectWrapper:
        project = self.conn.getObject('Project', project_id)
        return project

    def _get_dataset(self, dataset_id: int) -> omero.gateway.DatasetWrapper:
        dataset = self.conn.getObject('Dataset', dataset_id)
        return dataset

    def get_image_object(self, image_id: int) -> omero.gateway.ImageWrapper:
        image_object = self.conn.getObject('Image', image_id)
        return image_object

    def create_pixels_store(self, image_object: omero.gateway.ImageWrapper) -> omero.gateway.ProxyObjectWrapper:
        pixels_store = self.conn.createRawPixelsStore()
        pixels_store.setPixelsId(image_object.getPixelsId(), False, self.conn.SERVICE_OPTS)
        return pixels_store

    def get_annotation_image_ids(self) -> list:
        input_omero = self.params['input'].get('omero', {})
        include = input_omero['include']
        exclude = input_omero.get('exclude', {})
        target_labels = ensure_list(input_omero.get('labels', []))
        # include
        image_ids = set(ensure_list(include.get('image', [])))
        for dataset_id in ensure_list(include.get('dataset', [])):
            image_ids.update(self._get_dataset_annotation_image_ids(dataset_id, target_labels))
        for project_id in ensure_list(include.get('project', [])):
            project = self._get_project(project_id)
            for dataset in project.listChildren():
                image_ids.update(self._get_dataset_annotation_image_ids(dataset.getId(), target_labels))
        # exclude
        for id in ensure_list(exclude.get('image', [])):
            if id in image_ids:
                image_ids.remove(id)
        for dataset_id in ensure_list(exclude.get('dataset', [])):
            for id in self._get_dataset_annotation_image_ids(dataset_id):
                if id in image_ids:
                    image_ids.remove(id)
        for project_id in ensure_list(exclude.get('project', [])):
            project = self._get_project(project_id)
            for dataset in project.listChildren():
                for id in self._get_dataset_annotation_image_ids(dataset.getId()):
                    if id in image_ids:
                        image_ids.remove(id)
        return list(image_ids)

    def _get_dataset_annotation_image_ids(self, dataset_id: int, target_labels: list = []) -> list:
        dataset = self._get_dataset(dataset_id)
        image_ids = []
        for image_object in dataset.listChildren():
            annotations = self._get_image_annotations(image_object, target_labels)
            if len(annotations) == len(target_labels):
                image_ids.append(image_object.getId())
        return image_ids

    def get_image_annotation(self, image_id: int, target_labels: list) -> tuple:
        image_object = self.get_image_object(image_id)
        name = image_object.getName()
        annotations = self._get_image_annotations(image_object, target_labels)
        return name, annotations

    def _get_image_annotations(self, image_object: omero.gateway.ImageWrapper, annotation_keys: list) -> dict:
        annotations = {}
        for omero_annotation in image_object.listAnnotations():
            if omero_annotation.OMERO_TYPE == omero.model.MapAnnotationI:
                for annotation_key in annotation_keys:
                    for annotation in omero_annotation.getMapValue():
                        if annotation.name.lower() == annotation_key.lower():
                            annotations[annotation_key] = annotation.value
        return annotations

    def print_projects(self):
        projects = self.conn.listProjects()      # may include other users' data
        for project in projects:
            print_omero_object(project)


def print_omero_object(omero_object: omero.gateway.BlitzObjectWrapper, indent: int = 0):
    logging.info("""%s%s:%s  Name:"%s" (owner=%s)""" % (
        " " * indent,
        omero_object.OMERO_CLASS,
        omero_object.getId(),
        omero_object.getName(),
        omero_object.getOwnerOmeName()))

    for child in omero_object.listChildren():
        logging.info('\t', child.getName())



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from OmeSource import OmeSource

import Ice
import os
import omero.gateway
import omero.model
import toml
from uuid import uuid4

from src.OmeSliCC.image_util import *
from src.OmeSliCC.util import *
from src.OmeSliCC.XmlDict import dict2xml, XmlDict


name = toml.load("pyproject.toml")["project"]["name"]
version = toml.load("pyproject.toml")["project"]["version"]

# https://www.openmicroscopy.org/Schemas/Documentation/Generated/OME-2016-06/ome.html
OME_URI = "http://www.openmicroscopy.org/Schemas/OME/2016-06"
OME_XSI = "http://www.w3.org/2001/XMLSchema-instance"
OME_SCHEMA_LOC = f"{OME_URI} {OME_URI}/ome.xsd"


def create_ome_metadata(source: OmeSource,
                        output_filename: str,
                        combine_rgb: bool = True,
                        pyramid_sizes_add: list = None) -> str:

    file_name = os.path.basename(output_filename)
    file_title = get_filetitle(file_name)
    uuid = f'urn:uuid:{uuid4()}'

    ome = source.get_metadata().copy() if source.has_ome_metadata else XmlDict()
    ome['@xmlns'] = OME_URI
    ome['@xmlns:xsi'] = OME_XSI
    ome['@xsi:schemaLocation'] = OME_SCHEMA_LOC
    ome['@UUID'] = uuid
    ome['@Creator'] = f'{name} {version}'

    experimenter = ome.get('Experimenter')

    mag = source.get_mag()
    instrument = ome.get('Instrument')
    objective = ome.get('Instrument', {}).get('Objective')
    if mag != 0:
        if instrument is None:
            instrument = {'@ID': 'Instrument:0'}
        if objective is None:
            objective = {'@ID': 'Objective:0'}
            instrument['Objective'] = objective
        objective['@NominalMagnification'] = mag
        ome['Instrument'] = instrument

    # currently only supporting single image
    nimages = 1

    images = []
    for imagei in range(nimages):
        images0 = ensure_list(ome.get('Image', []))
        if len(images0) > 0:
            image0 = images0[imagei]
        else:
            image0 = {}
        pixels0 = image0.get('Pixels', {})

        channels = source.get_channels().copy()
        nchannels = source.get_size_xyzct()[3]
        if combine_rgb and len(channels) == 3:
            channel = channels[0].copy()
            channel['@SamplesPerPixel'] = nchannels
            channel.pop('Color', None)
            channels = [channel]
        elif not combine_rgb and len(channels) < nchannels:
            channel = channels[0].copy()
            channel['@SamplesPerPixel'] = 1
            channels = [channel] * nchannels

        for channeli, channel in enumerate(channels):
            channel['@ID'] = f'Channel:{imagei}:{channeli}'

        image = {
            '@ID': f'Image:{imagei}',
            '@Name': file_title,
        }

        xyzct = source.get_size_xyzct()
        pixels = {
            '@ID': f'Pixels:{imagei}',
            '@SizeX': xyzct[0],
            '@SizeY': xyzct[1],
            '@SizeZ': xyzct[2],
            '@SizeC': nchannels,
            '@SizeT': xyzct[4],
            '@Type': str(ensure_unsigned_type(source.get_pixel_type())),
            '@DimensionOrder': 'XYZCT',
            'Channel': channels,
            'TiffData': {'UUID': {'@FileName': file_name, '#text': uuid}},
        }

        pixel_size = source.get_pixel_size()
        if len(pixel_size) > 0 and pixel_size[0][0] != 0:
            pixels['@PhysicalSizeX'] = pixel_size[0][0]
        if len(pixel_size) > 1 and pixel_size[1][0] != 0:
            pixels['@PhysicalSizeY'] = pixel_size[1][0]
        if len(pixel_size) > 2 and pixel_size[2][0] != 0:
            pixels['@PhysicalSizeZ'] = pixel_size[2][0]
        if len(pixel_size) > 0 and pixel_size[0][1] != '':
            pixels['@PhysicalSizeXUnit'] = pixel_size[0][1]
        if len(pixel_size) > 1 and pixel_size[1][1] != '':
            pixels['@PhysicalSizeYUnit'] = pixel_size[1][1]
        if len(pixel_size) > 2 and pixel_size[2][1] != '':
            pixels['@PhysicalSizeZUnit'] = pixel_size[2][1]

        planes = ensure_list(pixels0.get('Plane', []))
        if len(planes) > 0:
            pixels['Plane'] = planes

        if 'AcquisitionDate' in image0:
            image['AcquisitionDate'] = image0['AcquisitionDate']
        if 'Description' in image0:
            image['Description'] = image0['Description']
        # Set image refs
        if experimenter is not None:
            image['ExperimenterRef'] = {'@ID': experimenter['ID']}
        if instrument is not None:
            image['InstrumentRef'] = {'@ID': instrument['ID']}
        if objective is not None:
            image['ObjectiveSettings'] = {'@ID': objective['ID']}
        # (end image refs)
        if 'StageLabel' in image0:
            image['StageLabel'] = image0['StageLabel']
        image['Pixels'] = pixels
        images.append(image)

    ome['Image'] = images

    if 'StructuredAnnotations' not in ome:
        ome['StructuredAnnotations'] = {}

    # filter source pyramid sizes
    map_annotations0 = ensure_list(ome['StructuredAnnotations'].get('MapAnnotation', []))
    map_annotations = [annotation for annotation in map_annotations0
                       if 'resolution' not in annotation.get('ID', '').lower()]
    # add pyramid sizes
    if pyramid_sizes_add is not None:
        key_value_map = {'M': [{'@K': i + 1, '#text': f'{" ".join([str(size) for size in pyramid_size])}'}
                               for i, pyramid_size in enumerate(pyramid_sizes_add)]}
        map_annotations.insert(0, {
            '@ID': 'Annotation:Resolution:0',
            '@Namespace': 'openmicroscopy.org/PyramidResolution',
            'Value': key_value_map
        })
    ome['StructuredAnnotations']['MapAnnotation'] = map_annotations

    # filter original metadata elements
    xml_annotations0 = ensure_list(ome.get('StructuredAnnotations', {}).get('XMLAnnotation', []))
    xml_annotations = [annotation for annotation in xml_annotations0
                       if 'originalmetadata' not in annotation.get('Namespace', '').lower()]
    ome['StructuredAnnotations']['XMLAnnotation'] = xml_annotations

    return dict2xml({'OME': ome})


def create_ome_metadata_from_omero(source: OmeSource,
                                   image_object: omero.gateway.ImageWrapper,
                                   output_filename: str,
                                   combine_rgb: bool = True,
                                   pyramid_sizes_add: list = None) -> str:

    file_name = os.path.basename(output_filename)
    file_title = get_filetitle(file_name)
    uuid = f'urn:uuid:{uuid4()}'

    ome = XmlDict()
    ome['@xmlns'] = OME_URI
    ome['@xmlns:xsi'] = OME_XSI
    ome['@xsi:schemaLocation'] = OME_SCHEMA_LOC
    ome['@UUID'] = uuid
    ome['@Creator'] = f'{name} {version}'

    instrument = None
    objective = None
    instrument0 = image_object.getInstrument()
    if instrument0 is not None:
        instrument = {'@ID': 'Instrument:0'}
        objectives0 = instrument0.getObjectives()
        if objectives0 is not None:
            objective0 = objectives0[0]
            objective = {
                '@ID': 'Objective:0',
                '@Manufacturer': objective0.getManufacturer(),
                '@Model': objective0.getModel(),
                '@LotNumber': objective0.getLotNumber(),
                '@SerialNumber': objective0.getSerialNumber(),
                '@NominalMagnification': source.get_mag(),
                '@CalibratedMagnification': source.get_mag(),
                '@LensNA': objective0.getLensNA(),
                '@WorkingDistance': objective0.getWorkingDistance().getValue(),
                '@WorkingDistanceUnit': objective0.getWorkingDistance().getSymbol(),
                '@Iris': objective0.getIris(),
            }
            correction = objective0.getCorrection().getValue()
            if correction is not None and correction.lower() not in ['', 'unknown']:
                objective['@Correction'] = correction
            immersion = objective0.getImmersion().getValue()
            if immersion is not None and immersion.lower() not in ['', 'unknown']:
                objective['@Immersion'] = immersion
            instrument['Objective'] = objective
        ome['Instrument'] = instrument

    # currently only supporting single image
    nimages = 1

    images = []
    for imagei in range(nimages):
        channels = []
        planes = []

        pixels0 = image_object.getPrimaryPixels()

        stage0 = image_object.getStageLabel()
        if stage0 is not None:
            for plane0 in pixels0.copyPlaneInfo():
                planes.append({
                    '@TheC': plane0.getTheC(),
                    '@TheT': plane0.getTheT(),
                    '@TheZ': plane0.getTheZ(),
                    'DeltaT': plane0.getDeltaT(),
                    'ExposureTime': plane0.getExposureTime(),
                    'PositionX': stage0.getPositionX().getValue(),
                    'PositionY': stage0.getPositionY().getValue(),
                    'PositionZ': stage0.getPositionZ().getValue(),
                    'PositionXUnit': stage0.getPositionX().getSymbol(),
                    'PositionYUnit': stage0.getPositionY().getSymbol(),
                    'PositionZUnit': stage0.getPositionZ().getSymbol(),
                })

        channelso = image_object.getChannels()
        for channeli, channelo in enumerate(channelso):
            channell = channelo.getLogicalChannel()
            light_path = channell.getLightPath()
            if light_path is None:
                light_path = {}
            channel = {
                '@ID': f'Channel:{imagei}:{channeli}',
                '@Name': channelo.getName(),
                '@Color': channelo.getColor().getInt(),
                '@EmissionWave': channelo.getEmissionWave(),
                '@ExcitationWave': channelo.getExcitationWave(),
                '@PockelCellSetting': channelo.getPockelCellSetting(),
                '@Fluor': channell.getFluor(),
                '@ContrastMethod': channell.getContrastMethod(),
                '@PinHoleSize': channell.getPinHoleSize(),
                '@SamplesPerPixel': channell.getSamplesPerPixel(),
                '@NDFilter': channell.getNdFilter(),
                'LightPath': light_path,
            }
            channels.append(channel)

        nchannels = source.get_size_xyzct()[3]
        if combine_rgb and len(channels) == 3:
            channel = channels[0].copy()
            channel['@SamplesPerPixel'] = nchannels
            channel.pop('Color', None)
            channels = [channel]
        elif not combine_rgb and len(channels) < nchannels:
            channel = channels[0].copy()
            channel['@SamplesPerPixel'] = 1
            channels = [channel] * nchannels

        image = {
            '@ID': f'Image:{imagei}',
            '@Name': file_title,
            'AcquisitionDate': image_object.getAcquisitionDate().isoformat(),
            'Description': image_object.getDescription(),
        }

        xyzct = source.get_size_xyzct()
        pixels = {
            '@ID': f'Pixels:{imagei}',
            '@SizeX': xyzct[0],
            '@SizeY': xyzct[1],
            '@SizeZ': xyzct[2],
            '@SizeC': nchannels,
            '@SizeT': xyzct[4],
            '@Type': str(ensure_unsigned_type(source.get_pixel_type())),
            '@DimensionOrder': 'XYZCT',
            'Channel': channels,
            'TiffData': {'UUID': {'@FileName': file_name, '#text': uuid}},
        }
        pixel_size = source.get_pixel_size()
        if len(pixel_size) > 0 and pixel_size[0][0] != 0:
            pixels['@PhysicalSizeX'] = pixel_size[0][0]
        if len(pixel_size) > 1 and pixel_size[1][0] != 0:
            pixels['@PhysicalSizeY'] = pixel_size[1][0]
        if len(pixel_size) > 2 and pixel_size[2][0] != 0:
            pixels['@PhysicalSizeZ'] = pixel_size[2][0]
        if len(pixel_size) > 0 and pixel_size[0][1] != '':
            pixels['@PhysicalSizeXUnit'] = pixel_size[0][1]
        if len(pixel_size) > 1 and pixel_size[1][1] != '':
            pixels['@PhysicalSizeYUnit'] = pixel_size[1][1]
        if len(pixel_size) > 2 and pixel_size[2][1] != '':
            pixels['@PhysicalSizeZUnit'] = pixel_size[2][1]

        if len(planes) > 0:
            pixels['Plane'] = planes

        # Set image refs
        if instrument is not None:
            image['InstrumentRef'] = {'@ID': instrument['@ID']}
        if objective is not None:
            image['ObjectiveSettings'] = {'@ID': objective['@ID']}
        # (end image refs)
        if stage0 is not None:
            image['StageLabel'] = {'@Name': stage0.getName()}
        image['Pixels'] = pixels
        images.append(image)

    ome['Image'] = images

    annotations = {}
    for annotations0 in image_object.listAnnotations():
        annotation_type = annotations0.OMERO_TYPE.__name__
        if annotation_type.endswith('I'):
            annotation_type = annotation_type[:-1]
        if annotation_type not in annotations:
            annotations[annotation_type] = []
        value = annotations0.getValue()
        if annotation_type == 'MapAnnotation':
            value = {'M': [{'@K': k, '#text': v} for k, v in value]}
        annotations[annotation_type].append({
            '@ID': f'Annotation:{len(annotations[annotation_type])}',
            '@Namespace': annotations0.getNs(),
            'Value': value
        })
    # add pyramid sizes
    if pyramid_sizes_add is not None:
        if 'MapAnnotation' not in annotations:
            annotations['MapAnnotation'] = []
        key_value_map = {'M': [{'@K': i + 1, '#text': f'{" ".join([str(size) for size in pyramid_size])}'}
                               for i, pyramid_size in enumerate(pyramid_sizes_add)]}
        annotations['MapAnnotation'].insert(0, {
            '@ID': 'Annotation:Resolution:0',
            '@Namespace': 'openmicroscopy.org/PyramidResolution',
            'Value': key_value_map
        })

    ome['StructuredAnnotations'] = annotations

    return dict2xml({'OME': filter_dict(ome)})


def get_omero_metadata_dict(image_object):
    # semi-automatically extract metadata dict (currently unused)
    ome = get_omero_metadata_dict_it(image_object)
    ome['Image'] = {
        '@ID': 'Image:0',
        '@Name': ome.pop('Name'),
        '@AcquisitionDate': ome.pop('AcquisitionDate'),
        'ObjectiveSettings': ome.pop('ObjectiveSettings'),
        'StageLabel': ome.pop('StageLabel'),
        'Pixels': ome.pop('Pixels')
    }
    annotations = {}
    for annotations0 in image_object.listAnnotations():
        annotation_type = annotations0.OMERO_TYPE.__name__
        if annotation_type.endswith('I'):
            annotation_type = annotation_type[:-1]
        if annotation_type not in annotations:
            annotations[annotation_type] = []
        value = annotations0.getValue()
        if annotation_type == 'MapAnnotation':
            value = {'M': [{'@K': k, '#text': v} for k, v in value]}
        annotations[annotation_type].append({
            '@ID': f'Annotation:{len(annotations[annotation_type])}',
            '@Namespace': annotations0.getNs(),
            'Value': value
        })
    ome['StructuredAnnotations'] = annotations
    return ome


def get_omero_metadata_dict_it(omero_obj, parents=[]):
    # recursively extract metadata dict (currently unused)
    metadata = {}
    if isinstance(omero_obj, omero.gateway.BlitzObjectWrapper):
        obj = omero_obj._obj
    else:
        obj = omero_obj
    for field in obj._field_info._fields:
        name = field[0].capitalize() + field[1:]
        if field != 'details' and field != 'relatedTo' \
                and not (name == 'Image' or name == 'Images' or name in parents) \
                and not ('ObjectiveSettings' in parents and name == 'Objective'):
            try:
                if hasattr(omero_obj, 'get' + name):
                    if name == 'Pixels':
                        value0 = omero_obj.getPrimaryPixels()
                    else:
                        value0 = getattr(omero_obj, 'get' + name)()
                    if value0 is not None:
                        data = []
                        for value in ensure_list(value0):
                            if isinstance(value, omero.gateway.BlitzObjectWrapper) or isinstance(value, Ice.Object):
                                if hasattr(value, 'getValue'):
                                    while hasattr(value, 'getValue'):
                                        value = value.getValue()
                                else:
                                    value = get_omero_metadata_dict_it(value, parents + [name])
                                data.append(value)
                            else:
                                data.append(value)
                        if len(data) > 0:
                            if len(data) == 1:
                                data = data[0]
                            metadata[name] = data
            except:
                pass
    return metadata



from getpass import getpass
import rsa


def generate_asymmetric_keys(private_key_filename: str, public_key_filename: str):
    pub_key, pri_key = rsa.newkeys(2048)
    with open(public_key_filename, 'w') as pub_file, open(private_key_filename, 'w') as pri_file:
        pub_file.write(pub_key.save_pkcs1().decode())
        pri_file.write(pri_key.save_pkcs1().decode())


def create_credentials_file(public_key_filename: str, credentials_filename: str):
    with open(public_key_filename, 'r') as key_file, open(credentials_filename, 'wb') as enc_file:
        keydata = key_file.read().encode()
        pub_key = rsa.PublicKey.load_pkcs1(keydata)
        enc_cred = rsa.encrypt((input('Username: ') + '\t' + getpass('Password: ')).encode(), pub_key)
        enc_file.write(enc_cred)
        print('credentials file generated')


def decrypt_credentials(private_key_filename: str, credentials_filename: str) -> tuple:
    with open(credentials_filename, 'rb') as cred_f, open(private_key_filename, 'rb') as key_file:
        keydata = key_file.read()
        pri_key = rsa.PrivateKey.load_pkcs1(keydata)
        cred = cred_f.read()
        dec_cred = rsa.decrypt(cred, pri_key).decode().split()
        usr = dec_cred[0]
        pwd = dec_cred[1]
    return usr, pwd



import io
import numpy as np
import PIL.Image
import omero.gateway

from src.OmeSliCC import Omero
from src.OmeSliCC.OmeSource import OmeSource
from src.OmeSliCC.ome import create_ome_metadata_from_omero
from src.OmeSliCC.XmlDict import XmlDict
from src.OmeSliCC.util import *


class OmeroSource(OmeSource):
    """Omero image source"""

    omero: Omero
    """Omero instance"""
    image_id: int
    """Omero image id"""
    image_object: omero.gateway.ImageWrapper
    """Omero image object"""
    pixels_store: omero.gateway.ProxyObjectWrapper
    """Raw pixels store object"""
    pixels_store_pyramid_order: list
    """Raw pixels store pyramid sizes order (pixel store level order not guaranteed) """

    def __init__(self,
                 omero: Omero,
                 image_id: int,
                 source_pixel_size: list = None,
                 target_pixel_size: list = None,
                 source_info_required: bool = False):

        super().__init__()
        self.omero = omero
        image_object = self.omero.get_image_object(image_id)
        self.image_object = image_object

        zsize = get_default(image_object.getPixelSizeZ(), 1)
        nchannels = np.sum([channel.getLogicalChannel().getSamplesPerPixel() for channel in image_object.getChannels()])
        pixel_type = np.dtype(image_object.getPixelsType())
        self.pixels_store = self.omero.create_pixels_store(self.image_object)
        for resolution in self.pixels_store.getResolutionDescriptions():
            self.sizes.append((resolution.sizeX, resolution.sizeY))
            self.sizes_xyzct.append((resolution.sizeX, resolution.sizeY, zsize, nchannels, 1))
            self.pixel_types.append(pixel_type)
            self.pixel_nbits.append(pixel_type.itemsize * 8)

        # Omero API issue: pixel store level order not guaranteed
        default_level = self.pixels_store.getResolutionLevel()
        if default_level != 0:
            # reverse order
            self.pixels_store_pyramid_order = list(reversed(range(default_level + 1)))
        else:
            # default order
            self.pixels_store_pyramid_order = list(range(len(self.sizes)))

        self._init_metadata(str(image_id),
                            source_pixel_size=source_pixel_size,
                            target_pixel_size=target_pixel_size,
                            source_info_required=source_info_required)

    def _find_metadata(self):
        #metadata = get_omero_metadata_dict(self.image_object)
        image_object = self.image_object
        self.source_pixel_size = [(get_default(image_object.getPixelSizeX(), 0), 'µm'),
                                  (get_default(image_object.getPixelSizeY(), 0), 'µm'),
                                  (get_default(image_object.getPixelSizeZ(), 0), 'µm')]
        self.source_mag = image_object.getObjectiveSettings().getObjective().getNominalMagnification()
        self.channels = []
        for channel in image_object.getChannels():
            channel = XmlDict({'@Name': channel.getName(), '@Color': channel.getColor().getInt(),
                               '@SamplesPerPixel': channel.getLogicalChannel().getSamplesPerPixel()})
            self.channels.append(channel)

    def create_xml_metadata(self, output_filename: str, combine_rgb: bool = True, pyramid_sizes_add: list = None) -> str:
        return create_ome_metadata_from_omero(self, self.image_object, output_filename, combine_rgb=combine_rgb,
                                              pyramid_sizes_add=pyramid_sizes_add)

    def get_thumbnail(self, target_size: tuple, precise: bool = False) -> np.ndarray:
        image_bytes = self.image_object.getThumbnail(target_size)
        image_stream = io.BytesIO(image_bytes)
        image = np.array(PIL.Image.open(image_stream))
        return image

    def _asarray_level(self, level: int, x0: float = 0, y0: float = 0, x1: float = -1, y1: float = -1) -> np.ndarray:
        pixels_store = self.pixels_store
        w, h = x1 - x0, y1 - y0
        nchannels = self.sizes_xyzct[level][3]
        image = np.zeros((h, w, nchannels), dtype=self.pixel_types[level])
        pixels_store.setResolutionLevel(self.pixels_store_pyramid_order[level])
        for c in range(nchannels):
            tile0 = pixels_store.getTile(0, c, 0, x0, y0, w, h)
            tile = np.frombuffer(tile0, dtype=image.dtype)
            tile.resize(h, w)
            image[..., c] = tile
        if nchannels == 1:
            return image[..., 0]
        else:
            return image

    def close(self):
        self.pixels_store.close()



# https://pypi.org/project/tifffile/


import os
from enum import Enum
import numpy as np
from tifffile import TiffFile, TiffPage
from concurrent.futures import ThreadPoolExecutor

from src.OmeSliCC import XmlDict
from src.OmeSliCC.OmeSource import OmeSource
from src.OmeSliCC.image_util import *
from src.OmeSliCC.util import *


class TiffSource(OmeSource):
    """Tiff-compatible image source"""

    filename: str
    """original filename"""
    loaded: bool
    """if image data is loaded"""
    decompressed: bool
    """if image data is decompressed"""
    pages: list
    """list of all relevant TiffPages"""
    data: bytes
    """raw un-decoded image byte data"""
    arrays: list
    """list of all image arrays for different sizes"""
    executor: ThreadPoolExecutor
    """ThreadPoolExecutor to be used for threaded operations"""

    def __init__(self,
                 filename: str,
                 source_pixel_size: list = None,
                 target_pixel_size: list = None,
                 source_info_required: bool = False,
                 executor: ThreadPoolExecutor = None):

        super().__init__()
        self.loaded = False
        self.decompressed = False
        self.data = bytes()
        self.arrays = []

        if executor is not None:
            self.executor = executor
        else:
            max_workers = (os.cpu_count() or 1) + 4
            self.executor = ThreadPoolExecutor(max_workers)

        tiff = TiffFile(filename)
        if tiff.is_ome and tiff.ome_metadata is not None:
            xml_metadata = tiff.ome_metadata
            self.metadata = XmlDict.xml2dict(xml_metadata)
            if 'OME' in self.metadata:
                self.metadata = self.metadata['OME']
                self.has_ome_metadata = True
        elif tiff.is_imagej:
            self.metadata = tiff.imagej_metadata

        self.pages = get_tiff_pages(tiff)
        for page0 in self.pages:
            npages = len(page0)
            if isinstance(page0, list):
                page = page0[0]
            else:
                page = page0
            shape = page.shape
            if isinstance(page, TiffPage):
                width = page.imagewidth
                height = page.imagelength
                depth = page.imagedepth * npages
                bitspersample = page.bitspersample
            else:
                width = shape[1]
                height = shape[0]
                depth = npages
                bitspersample = page.dtype.itemsize * 8
            nchannels = shape[2] if len(shape) > 2 else 1
            nt = 1
            if tiff.is_ome:
                pixels = self.metadata.get('Image', {}).get('Pixels', {})
                depth = int(pixels.get('SizeZ', depth))
                nchannels = int(pixels.get('SizeC', nchannels))
                nt = int(pixels.get('SizeT', nt))
            self.sizes.append((width, height))
            self.sizes_xyzct.append((width, height, depth, nchannels, nt))
            self.pixel_types.append(page.dtype)
            self.pixel_nbits.append(bitspersample)

        self.fh = tiff.filehandle

        self._init_metadata(filename,
                            source_pixel_size=source_pixel_size,
                            target_pixel_size=target_pixel_size,
                            source_info_required=source_info_required)

    def _find_metadata(self):
        pixel_size = []
        pixel_size_unit = ''
        channels = []
        mag = 0
        page = self.pages[0]
        if isinstance(page, list):
            page = page[0]
        # from OME metadata
        if page.is_ome:
            self._get_ome_metadate()
            return

        # from imageJ metadata
        pixel_size_z = None
        if len(pixel_size) == 0 and self.metadata is not None and 'spacing' in self.metadata:
            pixel_size_unit = self.metadata.get('unit', '')
            pixel_size_z = (self.metadata['spacing'], pixel_size_unit)
        if mag == 0 and self.metadata is not None:
            mag = self.metadata.get('Mag', 0)
        # from page TAGS
        metadata = tags_to_dict(page.tags)
        if len(pixel_size) < 2:
            if pixel_size_unit == '':
                pixel_size_unit = metadata.get('ResolutionUnit', '')
                if isinstance(pixel_size_unit, Enum):
                    pixel_size_unit = pixel_size_unit.name
                pixel_size_unit = pixel_size_unit.lower()
                if pixel_size_unit == 'none':
                    pixel_size_unit = ''
            res0 = metadata.get('XResolution')
            if res0 is not None:
                if isinstance(res0, tuple):
                    res0 = res0[0] / res0[1]
                if res0 != 0:
                    pixel_size.append((1 / res0, pixel_size_unit))
            res0 = metadata.get('YResolution')
            if res0 is not None:
                if isinstance(res0, tuple):
                    res0 = res0[0] / res0[1]
                if res0 != 0:
                    pixel_size.append((1 / res0, pixel_size_unit))
        if len(channels) == 0:
            nchannels = self.sizes_xyzct[0][3]
            photometric = str(metadata.get('PhotometricInterpretation', '')).lower().split('.')[-1]
            channels = [XmlDict.XmlDict({'@Name': photometric, '@SamplesPerPixel': nchannels})]
        if mag == 0:
            mag = metadata.get('Mag', 0)
        # from description
        if not page.is_ome:
            metadata = desc_to_dict(page.description)
            if mag == 0:
                mag = metadata.get('Mag', metadata.get('AppMag', 0))
            if len(pixel_size) < 2 and 'MPP' in metadata:
                pixel_size.append((metadata['MPP'], 'µm'))
                pixel_size.append((metadata['MPP'], 'µm'))
        if pixel_size_z is not None and len(pixel_size) == 2:
            pixel_size.append(pixel_size_z)
        self.source_pixel_size = pixel_size
        self.source_mag = mag
        self.channels = channels

    def load(self, decompress: bool = False):
        self.fh.seek(0)
        self.data = self.fh.read()
        self.loaded = True
        if decompress:
            self.decompress()

    def unload(self):
        self.loaded = False
        del self.data
        self.clear_decompress()

    def decompress(self):
        self.clear_decompress()
        for page in self.pages:
            if isinstance(page, list):
                array = []
                for page1 in page:
                    array.append(self._decompress_page(page1))
                array = np.asarray(array)
            else:
                array = self._decompress_page(page)
            self.arrays.append(array)
        self.decompressed = True

    def clear_decompress(self):
        self.decompressed = False
        for array in self.arrays:
            del array
        self.arrays = []

    def _decompress_page(self, page: TiffPage) -> np.ndarray:
        pw = page.shape[1]
        ph = page.shape[0]
        array = np.zeros(page.shape, page.dtype)
        tile_width = page.tilewidth
        tile_height = page.tilelength
        x, y = 0, 0
        index = 0
        for offset, bytecount in zip(page.dataoffsets, page.databytecounts):
            if bytecount > 0:
                data = self.data[offset:offset + bytecount]
                if page.jpegtables is not None:
                    decoded = page.decode(data, index, page.jpegtables)
                else:
                    decoded = page.decode(data, index)
                tile = decoded[0]
                dw = tile.shape[-2]
                dh = tile.shape[-3]
                if x + dw > pw:
                    dw = pw - x
                if y + dh > ph:
                    dh = ph - y
                array[y:y + dh, x:x + dw, :] = tile[0, 0:dh, 0:dw, :]
            x += tile_width
            if x >= page.imagewidth:
                x = 0
                y += tile_height
            index += 1
        #self.decode(page, page.dataoffsets, page.databytecounts, tile_width, tile_height, nx, array)  # numpy is not thread-safe!
        return array

    def _asarray_level(self, level: int, x0: float = 0, y0: float = 0, x1: float = -1, y1: float = -1) -> np.ndarray:
        # based on tiffile asarray
        if x1 < 0 or y1 < 0:
            x1, y1 = self.sizes[level]

        if self.decompressed:
            array = self.arrays[level]
            return array[y0:y1, x0:x1]

        dw = x1 - x0
        dh = y1 - y0
        page0 = self.pages[level]
        page = page0[0] if isinstance(page0, list) else page0
        size_xyzct = self.sizes_xyzct[level]
        n = size_xyzct[2] * size_xyzct[3]

        tile_height, tile_width = page.chunks[0], page.chunks[1]
        tile_y0, tile_x0 = y0 // tile_height, x0 // tile_width
        tile_y1, tile_x1 = np.ceil([y1 / tile_height, x1 / tile_width]).astype(int)
        w = (tile_x1 - tile_x0) * tile_width
        h = (tile_y1 - tile_y0) * tile_height
        tile_per_line = int(np.ceil(page.imagewidth / tile_width))

        out = np.zeros((h, w, n), page.dtype)

        dataoffsets = []
        databytecounts = []
        tile_locations = []
        for i, page in enumerate(page0):
            for y in range(tile_y0, tile_y1):
                for x in range(tile_x0, tile_x1):
                    index = int(y * tile_per_line + x)
                    if index < len(page.databytecounts):
                        offset = page.dataoffsets[index]
                        count = page.databytecounts[index]
                        if count > 0:
                            dataoffsets.append(offset)
                            databytecounts.append(count)
                            target_y = (y - tile_y0) * tile_height
                            target_x = (x - tile_x0) * tile_width
                            tile_locations.append((target_y, target_x, i))

            self._decode(page, dataoffsets, databytecounts, tile_locations, out)

        target_y0 = y0 - tile_y0 * tile_height
        target_x0 = x0 - tile_x0 * tile_width
        image = out[target_y0: target_y0 + dh, target_x0: target_x0 + dw]
        if n == 1:
            return image[..., 0]
        else:
            return image

    def _decode(self, page: TiffPage, dataoffsets: list, databytecounts: list, tile_locations: list, out: np.ndarray):
        def process_decoded(decoded, index, out=out):
            segment, indices, shape = decoded
            y, x, i = tile_locations[index]
            _, h, w, n = shape
            # Note: numpy is not thread-safe
            out[y: y + h, x: x + w, i: i + n] = segment[0]

        for _ in self._segments(
                func=process_decoded,
                page=page,
                dataoffsets=dataoffsets,
                databytecounts=databytecounts
        ):
            pass

    def _segments(self, func: callable, page: TiffPage, dataoffsets: list, databytecounts: list) -> tuple:
        # based on tiffile segments
        def decode(args, page=page, func=func):
            decoded = page.decode(*args, jpegtables=page.jpegtables)
            return func(decoded, args[1])

        tile_segments = []
        for index in range(len(dataoffsets)):
            offset = dataoffsets[index]
            bytecount = databytecounts[index]
            if self.loaded:
                segment = self.data[offset:offset + bytecount]
            else:
                fh = page.parent.filehandle
                fh.seek(offset)
                segment = fh.read(bytecount)
            tile_segments.append((segment, index))
        yield from self.executor.map(decode, tile_segments, timeout=10)
